#include <string>
#include serial

#define XON 0x02
#define XOFF 0x03

#define WEIGHT_GROSS 'B'
#define WEIGHT_NET   'D'

/**
  DataBuff - Массив из которого извлекаем checksum
  DataLen - Длина массива для проверки выхода за границы.
  Offset - Сдвиг от начала массива, с которого начинаем читать checksum
  EndByte - Байт до которого будем читать.
*/
stock GetCheckSum(DataBuff[], DataLen, Offset, EndByte)
{
  new CheckSum = 0;

  for (new BuffIndex = 0; 
    (BuffIndex < EndByte) && (Offset + BuffIndex) < DataLen; 
    BuffIndex++)
  {
    CheckSum ^= DataBuff[Offset + BuffIndex];
  }

  return CheckSum & 0xFF;
}

// %02x - вывод не менее 2 цифр, добавьте к ним 0, если меньше.
stock NumToStr(BuffOut[], Number)
{
  strformat(BuffOut, 3, false, "%02x", Number);
  return 2;
}

/**
  DataBuff - Массив из которого вычитываем SymbolsCount байтов 
  для преобразования в hex число.
  DataLen - Длина массива для проверки выхода за границы.
  BufferIndex - Кол-во обработанных байтов.
 */
stock StrToHex(DataBuff[], DataLen, SymbolsCount, &BufferIndex)
{
  new Value = 0;

  for (new Counter = 0; 
    Counter < SymbolsCount && BufferIndex < DataLen;
    Counter++)
  {
    if ('0' <= DataBuff[BufferIndex] <= '9')
    {
      Value = (Value * 16) + (DataBuff[BufferIndex] - '0');
    }
    else if ('A' <= DataBuff[BufferIndex] <= 'F')
    {
      Value = (Value * 16) + (DataBuff[BufferIndex] - 'A' + 10);
    }
    BufferIndex++;
  }

  return Value;
}

/**
  DataBuff - Массив из которого вычитываем SymbolsCount байтов 
  для преобразования в dec число.
  DataLen - Длина массива для проверки выхода за границы.
  BufferIndex - Кол-во обработанных байтов.
 */
stock StrToDec(DataBuff[], DataLen, SymbolsCount, &BufferIndex)
{
  new Value = 0;
  
  for (new Counter = 0;
    Counter < SymbolsCount && BufferIndex < DataLen; 
    Counter++)
  {
    Value = (Value * 10) + (DataBuff[BufferIndex] - '0');
    BufferIndex++;
  }

  return Value;
}

// Проверка на соответствие ожидаемой контрольной суммы и фактической.
stock bool:AssertCheckSum(DataBuff[], DataLen, Offset, End, ExpectedCheckSum)
{
  new CurrentCheckSum = GetCheckSum(DataBuff, DataLen, Offset, End);
  if (ExpectedCheckSum != CurrentCheckSum)
  {
    printf("Checksum mismatch.");
    printf("CurrentCheckSum: %d, Expected: %d", CurrentCheckSum, ExpectedCheckSum);
    return false;
  }
  return true;
}

//  Парсинг данных в периодичном режиме.
stock bool:ParsePeriodic(DataBuff[], DataLen, &Weight, &Pow)
{
  new Value = 0;
  new bool:IsNegative = false;
  new BufferIndex = 0;
  new XorIndexBegin = 0;
  new XorEnd = 0;
  new ExpectedXorValue = 0;

  if (DataLen < 12) // ожидаем длину сообщения не меньше 12 байтов.
  {
    printf("Err. Expected Len = %d, Current Len=%d", 12, DataLen);
    return false;
  }
  
  if (DataBuff[BufferIndex++] != XON) // Проверка байта на признак наличия стартового байта.
  {
    printf("Incorrect start data");
    return false;
  }
  
  XorIndexBegin = BufferIndex;

  if (DataBuff[BufferIndex++] == 0x2D) IsNegative = true;
    
  /**
    Преобразование [2; 8) = 6 байтов в dec число, 
    которое, согласно протоколу [3; 9), является весом.
   */
  Value = StrToDec(DataBuff, DataLen, 6, BufferIndex); 

  if (IsNegative) Value *= -1;
  Weight = Value;

  // Получение значение Pow, в протоколе DecimalDigit(9-байт).
  Pow = (DataBuff[BufferIndex++]) - '0';

  // Байт, до которого будем парсить значение текущей CheckSum
  XorEnd = BufferIndex - 1;

  /**
    Получим ожидаемую CheckSum из текущего сообщения.
    Согласно протоколу [10; 11] байт.
   */
  ExpectedXorValue = StrToHex(DataBuff, DataLen, 2, BufferIndex);

  //  Проверим совпадает ли ожидаемая и текущая CheckSum.
  return AssertCheckSum(DataBuff, DataLen, XorIndexBegin, XorEnd, ExpectedXorValue);
}

stock bool:ParseAnswer(DataBuff[], DataLen, WeightType = WEIGHT_NET, &Weight, &Pow)
{
  new Value = 0;
  new bool:IsNegative = false;
  new BufferIndex = 0;
  new XorIndexBegin = 0;
  new XorEnd = 0;
  new ExpectedXorValue = 0;
  
  if (DataLen < 14) // ожидаем длину сообщения не меньше 14 байтов.
  {
    printf("Err. Expected Len = %d, Current Len=%d", 14, DataLen);
    return false;
  }
  if (DataBuff[BufferIndex++] != XON) // Проверка байта на признак наличия стартового байта.
  {
    printf("Incorrect start data.");
    return false;
  }

  XorIndexBegin = BufferIndex; 

  BufferIndex++; // Пропускаем Address Serial No

  if (DataBuff[BufferIndex++] != WeightType)
  {
    printf("Incorrect weight type.");
    return false;
  }
  if (DataBuff[BufferIndex++] == '-') IsNegative = true;

  // Преобразование [4; 10) = 6 байтов в dec число 
  Value = StrToDec(DataBuff, DataLen, 6, BufferIndex);
  if (IsNegative) Value *= -1;
  Weight = Value;

  Pow = (DataBuff[BufferIndex++]) - '0';
  XorEnd = BufferIndex - 1;

  ExpectedXorValue = StrToHex(DataBuff, DataLen, 2, BufferIndex);

  return AssertCheckSum(DataBuff, DataLen, XorIndexBegin, XorEnd, ExpectedXorValue);
}

// BuffOut - Выходной буффер
// SizeBuffOut - Длина массива для проверки выхода за границы.
stock ReadPeriodic(SerialPort:port = SERIAL_PORT_0, BuffOut[], SizeBuffOut)
{
  new Len = 0;
  new DataIndex = 0;
  new BuffRead[12 * 3];
  new BuffOutIndex = 0;

  Len = rsrecv(BuffRead, _, 1000, 100, port);

  // Ищем начало данных по XON байту.
  for (new Offset = 0; Offset < Len; Offset++)
  {
    if (BuffRead[Offset] == XON)
    {
      DataIndex = Offset;
      break;
    }
  }

  // BuffOut - буфер на выходе содержит сообщение без мусора.
  while ( (BuffOutIndex < SizeBuffOut) && (DataIndex < Len) )
  {
    BuffOut[BuffOutIndex++] = BuffRead[DataIndex++];
    if (BuffRead[DataIndex] == XOFF) 
    {
      BuffOut[BuffOutIndex] = BuffRead[DataIndex];
      BuffOutIndex++;
      break;
    }
  }

  printf("Data len=%d", BuffOutIndex);
  return BuffOutIndex; // return length
}

stock bool:ReadAnswer(SerialPort:port = SERIAL_PORT_0, WeightType = WEIGHT_NET, &Weight, &Pow)
{
  new Buff[32];
  new BuffIndex = 0;
  new Len = 0;
  new CheckSum = 0;
  
  // Формируем запрос
  Buff[BuffIndex++] = XON; 
  Buff[BuffIndex++] = 'A';
  Buff[BuffIndex++] = WeightType; // Запрашиваемый тип веса
  CheckSum = GetCheckSum(Buff, sizeof(Buff), 1, 2);
  BuffIndex += NumToStr(Buff[BuffIndex], CheckSum);
  Buff[BuffIndex++] = XOFF;

  rssend(Buff, BuffIndex, port); // Send request
  Len = rsrecv(Buff, 32, 1000, 100, port); // Read response
  printf("Read len = %d", Len);

  return ParseAnswer(Buff, Len, WeightType, Weight, Pow);
}
